/*
 * Copyright (C) 2001 USAGI/WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $USAGI: sadb.c,v 1.9 2001/08/15 06:30:29 mk Exp $
 */

/*
 * sadb.c is a program for IPsec SADB. It provide functions for SADB manipulation.
 * struct ipsec_sa represents IPsec SA. It has a reference count and maneges itself.
 * If you get a reference(pointer) of struct ipsec_sa by sadb_find_by_***, please
 * call ipsec_sa_release_ref when you abondan the reference. 
 */ 


#ifdef __KERNEL__

#ifdef MODULE
#  include <linux/module.h>
#  ifdef MODVERSIONS
#    include <linux/modversions.h>
#  endif /* MODVERSIONS */
#endif /* MODULE */

#ifdef SADB_DEBUG
#define DEBUG 1
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/malloc.h>
#include <linux/spinlock.h>

#include <net/pfkeyv2.h>
#include <net/sadb.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif /* CONFIG_PROC_FS */

#endif /* __KERNEL__ */

#include "sadb_utils.h"

#define BUFSIZE 64

#if 0
LIST_HEAD(sadb_larval_list);
LIST_HEAD(sadb_mature_list);
LIST_HEAD(sadb_dying_list);
LIST_HEAD(sadb_dead_list);
#endif


/*
 * sa_list represents SADB as a list. 
 * sadb_lock is a lock for SADB.
 * If you access a element in SADB,
 * you must lock and unlock each element before access.
 *
 * struct list_head *pos = NULL;
 * struct ipsec_sa *tmp = NULL;
 *
 * read_lock(&sadb_lock);
 * list_for_each(pos, &sadb_lock){
 *	tmp = list_entry(pos, struct ipsec_sa, entry);
 *	read_lock(&(tmp->lock));
 *
 *	some operation
 *
 *	read_unlock(&(tmp->lock));
 * }
 * read_unlock(&sadb_lock);
 *
 *  * ## If you want to change the list or a element,
 *    you read these code with write_lock, write_unlock
 *    instead of read_lock, read_unlock.
 */



LIST_HEAD(sadb_list);
rwlock_t sadb_lock = RW_LOCK_UNLOCKED;

struct ipsec_sa* ipsec_sa_kmalloc()
{
	struct ipsec_sa* sa = NULL;

	sa = (struct ipsec_sa*) kmalloc(sizeof(struct ipsec_sa), GFP_KERNEL);

	if (!sa) {
		pr_debug("sadb_kmalloc: sa couldn\'t be allocated\n");
		return NULL;
	}

	ipsec_sa_init(sa);

	return sa;
}

int ipsec_sa_init(struct ipsec_sa *sa)
{
	int error = 0;

	if (!sa) {
		pr_debug("sadb_kmalloc: sa is null\n");
		error = -EINVAL;
		goto err;
	}

	memset(sa, 0, sizeof(struct ipsec_sa));
	/* set default replay window size (32) -mk */
	sa->replay_window.size = 32;
	atomic_set(&(sa->count), 1);
	sa->lock = RW_LOCK_UNLOCKED;
err:
	return error;
}

int ipsec_sa_kfree(struct ipsec_sa **sa)
{
	int error = 0;
	if (!*sa) {
		pr_debug("sadb_kfree: sa is null\n");
		error = -EINVAL;
		goto err;
	}

	if (atomic_read(&(*sa)->count)) {
		pr_debug("sadb_kfree: sa has been referenced.\n");
		error = -EINVAL;
		goto err;
	}

	kfree(*sa);

err:
	return error;
}

int ipsec_sa_copy(struct ipsec_sa *dst, struct ipsec_sa *src)
{
	int error = 0;
	if (!(src&&dst)) {
		pr_debug("sadb_copy: src or dst is null\n");
		error = -EINVAL;
		goto err;
	}

	memcpy( dst, src, sizeof(struct ipsec_sa));
	dst->esp_algo.cx.keyinfo =  dst->esp_algo.key;

	atomic_set(&(dst->count),1);

err:
	return error;
}

int ipsec_sa_release_ref(struct ipsec_sa *sa)
{
	int error = 0;
	if (!sa) {
		pr_debug("sadb_release_ref: sa is null\n");
		error = -EINVAL;
		goto err;
	}

	atomic_dec(&(sa->count));
	pr_debug("ipsec_sa_release_ref: prt=%p,count=%d\n",
		sa, atomic_read(&(sa->count))); 
err:
	return error;
}

int sadb_append(struct ipsec_sa *sa)
{
	int error = 0;
	struct ipsec_sa *new;

	if (!sa) {
		pr_debug("sadb_append: sa == null.\n"); 
		error = -EINVAL;
		goto err;
	}

	error = sadb_find_by_address_proto_spi((struct sockaddr*)&(sa->src),
					 (struct sockaddr*)&(sa->dst),
					 (struct sockaddr*)&(sa->pxy),
					 sa->ipsec_proto,
					 sa->spi,
					 NULL);

	if (error == -EEXIST) {
		pr_debug("sadb_append: sa already exist\n");
		goto err;
	} else if (error == -EINVAL) {
		pr_debug("sadb_append: invalid parameters\n");
	}

	new = ipsec_sa_kmalloc();
	if (!new) {
		pr_debug("sadb_append: ipsec_sa_kmalloc failed\n");
		error = -ENOMEM;
		goto err;
	}
	error = ipsec_sa_copy(new, sa);
	if (error) {
		pr_debug("sadb_append: ipsec_sa_copy failed\n");
		goto err;
	}

	write_lock(&sadb_lock);
	list_add(&(new->entry), &sadb_list);
	write_unlock(&sadb_lock);

err:
	return error;
}

int sadb_remove(struct sockaddr *src, struct sockaddr *dst, __u32 spi)
{
	int error = -ESRCH;
	struct list_head *pos = NULL;
	struct ipsec_sa *tmp = NULL;
	char buf[BUFSIZE];

	if (!(src && dst)) {
		pr_debug("sadb_remove: src or dst is null.\n");
		error = -EINVAL;
		goto err;
	}

#ifdef SADB_DEBUG
	memset(buf, 0, BUFSIZE);
	sockaddrtoa(src, buf, BUFSIZE);
	pr_debug("sadb_remove: src=%s\n", buf);
	memset(buf, 0, BUFSIZE);
	sockaddrtoa(dst, buf, BUFSIZE);
	pr_debug("sadb_remove: dst=%s\n", buf);
	pr_debug("sadb_remove: spi=%d\n", ntohl(spi));
#endif
	write_lock(&sadb_lock);
	list_for_each(pos, &sadb_list){
		tmp = list_entry(pos, struct ipsec_sa, entry);
		if (tmp->spi == spi) {
			if (!compare_address(src, (struct sockaddr*)&(tmp->src)) &&
				!compare_address(dst, (struct sockaddr*)&(tmp->dst))) {
				pr_debug("sadb_remove: I found a sa to be removed.\n");
				if (atomic_dec_and_test(&(tmp->count))) {
					pr_debug("sadb_remove: sa=%p is removed\n", tmp);
					list_del(&(tmp->entry));
					kfree(tmp);
					error = 0;
					break;
				} else {
					pr_debug("sadb_remove: sa is refered by anyone.\n");
				}
			}
		}
	}
	write_unlock(&sadb_lock);

err:
	return error;
}

int sadb_update(struct ipsec_sa *entry)
{
	int error = 0;
	if (entry) {
		pr_debug("sadb_update entry == null\n");
		error = -EINVAL;
		goto err;
	}

	/* TODO: Current code removes old sa and append new sa,
		 but it is correct to update sa returnd from sa_find_by_said. */

	error = sadb_remove((struct sockaddr*)&(entry->src),
				(struct sockaddr*)&(entry->dst), entry->spi);
	if (error != -ESRCH) {
		pr_debug("sadb_update: I could not remove\n");
		goto err;
	}
	error = sadb_append(entry);
	if (error) {
		pr_debug("sadb_update: I could not append\n");
		goto err;
	}

err:
	return error;
}

int sadb_find_by_spi(__u32 spi, struct ipsec_sa **sa)
{
	int error = -ESRCH;
	struct list_head *pos;
	struct ipsec_sa *tmp;

	if (!sa) {
		pr_debug("sadb_find_by_spi: sa == null\n");
		error = -EINVAL;
		goto err;
	}

	read_lock(&sadb_lock);
	list_for_each(pos, &sadb_list){
		tmp = list_entry(pos, struct ipsec_sa, entry);
		read_lock(&(tmp->lock));
		if (tmp->spi == spi) {
			if (sa) {
				atomic_inc(&(tmp->count));
				*sa = tmp;
				pr_debug("sadb_find_by_spi: ptr=%p,count=%d\n",
					tmp, atomic_read(&(tmp->count)));
			}
			error = -EEXIST;
			read_unlock(&(tmp->lock));
			break;
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&sadb_lock);

err:
#ifdef SADB_DEBUG
	if (error && (error != -EEXIST))
		pr_debug("sadb_find_by_spi: I could not find any sa.\n");
#endif
	return error;
}

int sadb_find_by_said( struct sockaddr *dst,
		 	int ipsec_proto, __u32 spi,
			 struct ipsec_sa **sa)
{
	int error = -ESRCH;
	struct list_head *pos = NULL;
	struct ipsec_sa *tmp = NULL;
	char buf[BUFSIZE];

	if (!dst) {
		pr_debug("sadb_find_by_said: dst == null\n");
		error = -EINVAL;
		goto err;
	}

#ifdef SADB_DEBUG
		memset(buf, 0, BUFSIZE);
		sockaddrtoa(dst, buf, BUFSIZE);
		pr_debug("sadb_find_by_said: dst=%s\n", buf);
		pr_debug("sadb_find_by_said: ipsec_proto=%d\n", ipsec_proto);
		pr_debug("sadb_find_by_said: spi==0x%x\n", ntohl(spi));
#endif
	read_lock(&sadb_lock);
	list_for_each(pos, &sadb_list){
		tmp = list_entry(pos, struct ipsec_sa, entry);
		read_lock(&(tmp->lock));
		if (tmp->spi == spi &&
			tmp->ipsec_proto == ipsec_proto &&
			!compare_address(dst, (struct sockaddr*)&(tmp->dst))) {
			pr_debug("sadb_find_by_said: found sa matching params.\n");
			if (sa) {
				atomic_inc(&(tmp->count));
				*sa = tmp;
				pr_debug("sadb_find_by_said: ptr=%p,count=%d\n",
					tmp, atomic_read(&(tmp->count)));
			}
			error = -EEXIST;
			read_unlock(&(tmp->lock));
			break;
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&sadb_lock);
#ifdef SADB_DEBUG
	if (error && (error != -EEXIST))
		pr_debug("sadb_find_by_said: I could not find any sa.\n");
#endif
		
err:
	return error;
}

int sadb_find_by_sa_index(struct sa_index *sa_index)
{
	int error = -ESRCH;
	struct list_head *pos;
	struct ipsec_sa *tmp = NULL;
	char buf[BUFSIZE];

	if (!sa_index) {
		pr_debug("sadb_find_by_sa_index: sa_index is null\n");
		error = -EINVAL;
		goto err;
	}

#ifdef SADB_DEBUG
		memset(buf, 0, BUFSIZE);
		sockaddrtoa((struct sockaddr*)&(sa_index->dst), 0, BUFSIZE);
		pr_debug("sadb_find_by_sa_index: sa_index->dst=%s\n", buf);
		pr_debug("sadb_find_by_sa_index: sa_index->ipsec_proto=%u\n", sa_index->ipsec_proto);
		pr_debug("sadb_find_by_sa_index: sa_index->spi=0x%x\n", ntohl(sa_index->spi));
#endif /* SADB_DEBUG */

	read_lock(&sadb_lock);
	list_for_each(pos, &sadb_list){
		tmp = list_entry(pos, struct ipsec_sa, entry);
		read_lock(&(tmp->lock));
		if (tmp->spi == sa_index->spi &&
			tmp->ipsec_proto == sa_index->ipsec_proto && 
			!compare_address((struct sockaddr*)&(sa_index->dst), (struct sockaddr*)&(tmp->dst))) {
			pr_debug("sadb_find_by_said: found sa matching params.\n");
			if (sa_index) {
				atomic_inc(&(tmp->count));
				sa_index->sa = tmp;
				pr_debug("sadb_find_by_said: ptr=%p,count=%d\n",
					tmp, atomic_read(&(tmp->count)));
			}
			error = -EEXIST;
			read_unlock(&(tmp->lock));
			break;
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&sadb_lock);

#ifdef SADB_DEBUG
	if (error && (error != -EEXIST))
		pr_debug("sadb_find_by_said: I could not find any sa.\n");
#endif
		
err:
	return error;
}

int sadb_find_by_address_proto_spi(struct sockaddr *src,
				struct sockaddr *dst,
				struct sockaddr *pxy,
				__u8 ipsec_proto,
				__u32 spi,
				struct ipsec_sa **sa)
{
	int error = -ESRCH;
	struct list_head *pos = NULL;
	struct ipsec_sa *tmp = NULL;
	char buf[BUFSIZE];

	if (!dst) {
		pr_debug("sadb_find_by_said: dst is null\n");
		error = -EINVAL;
		goto err;
	}


#ifdef SADB_DEBUG
	memset(buf, 0, BUFSIZE);
	sockaddrtoa(src, buf, BUFSIZE);
	pr_debug("sadb_find_by_address_proto_spi: src=%s\n", buf);
	memset(buf, 0, BUFSIZE);
	sockaddrtoa(dst, buf, BUFSIZE);
	pr_debug("sadb_find_by_address_proto_spi: dst=%s\n", buf);
#endif 	

	read_lock(&sadb_lock);
	list_for_each(pos, &sadb_list){
		tmp = list_entry(pos, struct ipsec_sa, entry);
		read_lock(&(tmp->lock));
		if (tmp->ipsec_proto == ipsec_proto && tmp->spi == spi ) {
			if (!compare_address(dst, (struct sockaddr*)&(tmp->dst)) &&
				!compare_address(src, (struct sockaddr*)&(tmp->src)) && 
					spi == tmp->spi) {

				if (pxy && compare_address(pxy, (struct sockaddr*)&(tmp->pxy))) {
					pr_debug("sadb_find_by_said: found sa matching params.\n");
					if (sa) {
						atomic_inc(&(tmp->count));
						*sa = tmp;
						pr_debug("sadb_find_by_address_proto_spi ptr=%p,count%d\n",
							tmp, atomic_read(&(tmp->count)));
					}
					read_unlock(&(tmp->lock));
					error = -EEXIST;
					break;
				} else {
					pr_debug("sadb_find_by_said: found sa matching params.\n");
					if (sa) {
						atomic_inc(&(tmp->count));
						*sa = tmp;
						pr_debug("sadb_find_by_address_proto_spi ptr=%p,count%d\n",
							tmp, atomic_read(&(tmp->count)));
					}
					read_unlock(&(tmp->lock));
					error = -EEXIST;
					break;
				}
			}
		}
		read_unlock(&(tmp->lock));
	}
	read_unlock(&sadb_lock);
err:

#ifdef SADB_DEBUG
	if (error && (error != -EEXIST))
		pr_debug("sadb_find_by_said: I could not find any sa.\n");
#endif
		
	return error;
}

void sadb_clear_db(void)
{
	struct list_head *list_pos = NULL;
	struct ipsec_sa *tmp = NULL;

	write_lock(&sadb_lock);
	list_for_each(list_pos, &sadb_list){
		/* TODO: check reference count */
		tmp = list_entry(list_pos, struct ipsec_sa, entry);
#ifdef SADB_DEBUG
		pr_debug("sadb_clear_db: dump sa start\n");
		dump_ipsec_sa(tmp);
		pr_debug("sadb_clear_db: dump sa end\n");
#endif
		list_del(&(tmp->entry));
		kfree(tmp);
	}
	write_unlock(&sadb_lock);
}

/* XXX allocations, bytes,... -mk */
void sadb_lifetime_scan(void)
{
	struct list_head *list_pos = NULL;
	struct ipsec_sa *tmp = NULL;
	time_t ctime = jiffies / HZ; /* current time */

	write_lock(&sadb_lock);
	list_for_each(list_pos, &sadb_list) {
		tmp = list_entry(list_pos, struct ipsec_sa, entry);
		write_lock(&(tmp->lock));
		if ( tmp->lifetime_s.addtime != 0 && ctime > (tmp->lifetime_s.addtime + tmp->init_time) ) {
			printk(KERN_INFO "SA soft lifetime(addtime) over !\n");
			/* XXX report the information to userland applications ? */
		}
		if ( tmp->lifetime_s.usetime != 0 && ctime > (tmp->lifetime_s.usetime + tmp->fuse_time) ) {
			printk(KERN_INFO "SA soft lifetime(usetime) over !\n");
			/* XXX report the information to userland applications ? */
		}
		if ( tmp->lifetime_h.addtime != 0 && ctime > (tmp->lifetime_h.addtime + tmp->init_time) ) {
			printk(KERN_INFO "SA hard lifetime(addtime) over! : ");
			if (atomic_dec_and_test(&(tmp->count))) {
				printk(KERN_INFO "sa=%p was removed\n", tmp);
				list_del(&(tmp->entry));
				kfree(tmp);
			} else {
				printk(KERN_INFO "sa is refered by anyone.\n");
			}
		}
		if ( tmp->lifetime_h.usetime != 0 && ctime > (tmp->lifetime_h.usetime + tmp->fuse_time) ) {
			printk(KERN_INFO "SA hard lifetime(usetime) over ! : ");
			if (atomic_dec_and_test(&(tmp->count))) {
				printk(KERN_INFO "sa=%p was removed\n", tmp);
				list_del(&(tmp->entry));
				kfree(tmp);
			} else {
				printk(KERN_INFO "sa is referd by anyone.\n");
			}
		}
		write_unlock(&(tmp->lock));
	}
	write_unlock(&sadb_lock);
}

#ifdef CONFIG_PROC_FS
static int sadb_get_info(char *buffer, char **start, off_t offset, int length )
{
        int len = 0;
        off_t pos=0;
        off_t begin=0;
	char buf[BUFSIZE]; 
	struct list_head *list_pos = NULL;
	struct ipsec_sa *tmp = NULL;
	len += sprintf(buffer + len, "---------------------------------------------\n"); 
	read_lock(&sadb_lock);
	list_for_each(list_pos, &sadb_list){
		tmp = list_entry(list_pos, struct ipsec_sa, entry);
		memset(buf, 0, BUFSIZE);
		sockaddrtoa((struct sockaddr*)&(tmp->src), buf, BUFSIZE);
		len += sprintf(buffer + len, "src:%s\n", buf);
		memset(buf, 0, BUFSIZE);
		sockaddrtoa((struct sockaddr*)&(tmp->dst), buf, BUFSIZE);
		len += sprintf(buffer + len, "dst:%s\n", buf);
		memset(buf, 0, BUFSIZE);
		sockaddrtoa((struct sockaddr*)&(tmp->pxy), buf, BUFSIZE);
		len += sprintf(buffer + len, "pxy:%s\n", buf);
		len += sprintf(buffer + len, "proto:%d spi:0x%X ", ntohl(tmp->proto), ntohl(tmp->spi) );
		len += sprintf(buffer + len, "auth:%d(%s) ", tmp->auth_algo.algo, authtoa(tmp->auth_algo.algo));
		len += sprintf(buffer + len, "esp :%d(%s)\n", tmp->esp_algo.algo, esptoa(tmp->esp_algo.algo));
		len += sprintf(buffer + len, "lifetime c:%u(byte), %u(add), %u(use)\n",
					tmp->lifetime_c.bytes, tmp->lifetime_c.addtime, tmp->lifetime_c.usetime);
		len += sprintf(buffer + len, "lifetime s:%u(byte), %u(add), %u(use)\n",
					tmp->lifetime_s.bytes, tmp->lifetime_s.addtime, tmp->lifetime_s.usetime);
		len += sprintf(buffer + len, "lifetime h:%u(byte), %u(add), %u(use)\n",
					tmp->lifetime_h.bytes, tmp->lifetime_h.addtime, tmp->lifetime_h.usetime);
		len += sprintf(buffer + len, "---------------------------------------------\n"); 

		pos=begin+len;
		if (pos<offset) {
			len=0;
			begin=pos;
		}
		if (pos>offset+length) {
			read_unlock(&sadb_lock);
			goto done;
		}
	}	
	read_unlock(&sadb_lock);
done:

        *start=buffer+(offset-begin);
        len-=(offset-begin);
        if (len>length)
                len=length;
        if (len<0)
                len=0;
        return len;
}
#endif /* CONFIG_PROC_FS */

int sadb_init(void)
{
	int error = 0;

#ifdef CONFIG_PROC_FS
	proc_net_create("sadb", 0, sadb_get_info);
#endif /* CONFIG_PROC_FS */

	pr_info("sadb_init: SADB initialized\n");
	return error;
}

int sadb_cleanup(void)
{
	int error = 0;

	INIT_LIST_HEAD(&sadb_list);
#ifdef CONFIG_PROC_FS
	proc_net_remove("sadb");
#endif /* CONFIG_PROC_FS */

	sadb_clear_db();

	pr_info("sadb_init: SADB cleaned up\n");
	return error;
}

#if 0

/* #ifdef MODULE */

static int __init sadb_module_init(void)
{
        int err = sadb_init();
        return err;
}

static void __exit sadb_module_cleanup(void)
{
        sadb_cleanup();
}

module_init(sadb_module_init);
module_exit(sadb_module_cleanup);


EXPORT_SYMBOL(ipsec_sa_release_ref);
EXPORT_SYMBOL(sadb_append);
EXPORT_SYMBOL(sadb_remove);
EXPORT_SYMBOL(sadb_update);
EXPORT_SYMBOL(sadb_clear_db);
EXPORT_SYMBOL(sadb_find_by_spi);
EXPORT_SYMBOL(sadb_find_by_said);
EXPORT_SYMBOL(sadb_find_by_sa_index);
EXPORT_SYMBOL(sadb_find_by_address_proto_spi);

#endif /* MODULE */


